/* eslint-disable @typescript-eslint/no-explicit-any */

/**
 * The contents in this file were taken from NorthernMan54/Hap-Node-Client
 * https://raw.githubusercontent.com/NorthernMan54/Hap-Node-Client/master/lib/httpParser.js
 */

// Borrowed and heaviliy modifed from https://github.com/miguelmota/http-message-parser

import { Buffer } from 'buffer';
import { Dictionary } from '../../Types/types';

interface IHttpResult {
    protocol?: string;
    httpVersion?: number;
    statusCode?: number;
    statusMessage?: string;
    method?: string;
    url?: string;
    headers?: Dictionary<string | number>;
    body?: any;
    boundary?: any;
    multipart?: any;
    additional?: any;
    meta?: any;
}

export default class HttpMessageParser {
    private requestLineRegex = /(HTTP|EVENT)\/(1\.0|1\.1|2\.0)\s+(\d+)\s+([\w\s-_]+)/i;
    private responseLineRegex = /(GET|POST|PUT|DELETE|PATCH|OPTIONS|HEAD|TRACE|CONNECT)\s+(.*)\s+HTTP\/(1\.0|1\.1|2\.0)/i;
    private headerNewlineRegex = /^[\r\n]+/gim;
    private boundaryRegex = /(\n|\r\n)+--[\w-]+(\n|\r\n)+/g;

    //eslint-disable-next-line
    public parse(message: any): IHttpResult {
        const result: IHttpResult = {};

        let messageString = '';
        let headerNewlineIndex = 0;
        //eslint-disable-next-line
        let fullBoundary: any = null;

        if (this.isBuffer(message)) {
            messageString = message.toString();
        } else if (typeof message === 'string') {
            messageString = message;
            message = new Buffer(messageString);
        } else {
            return result;
        }

        /*
         * Strip extra return characters
         */
        messageString = messageString.replace(/\r\n/gim, '\n');

        messageString = this.trim(messageString);

        /* Parse request line
        */
        const possibleRequestLine = messageString.split(/\n|\r\n/)[0];
        const requestLineMatch = possibleRequestLine.match(this.requestLineRegex);

        /* Parse request line
        */
        if (Array.isArray(requestLineMatch) && requestLineMatch.length > 1) {
            result.protocol = requestLineMatch[1];
            result.httpVersion = parseFloat(requestLineMatch[2]);
            result.statusCode = parseInt(requestLineMatch[3], 10);
            result.statusMessage = requestLineMatch[4];
        } else {
            const responseLineMath = possibleRequestLine.match(this.responseLineRegex);
            if (Array.isArray(responseLineMath) && responseLineMath.length > 1) {
                result.method = responseLineMath[1];
                result.url = responseLineMath[2];
                result.httpVersion = parseFloat(responseLineMath[3]);
            }
        }

        /* Parse headers
         */

        headerNewlineIndex = messageString.search(this.headerNewlineRegex);
        if (headerNewlineIndex > -1) {
            headerNewlineIndex = headerNewlineIndex + 1; // 1 for newline length
        } else {
            /* There's no line breaks so check if request line exists
             * because the message might be all headers and no body
             */
            if (result.httpVersion) {
                headerNewlineIndex = messageString.length;
            }
        }

        const headersString = messageString.substr(0, headerNewlineIndex);
        const headers = this.parseHeaders(headersString);

        if (Object.keys(headers).length > 0) {
            result.headers = headers;

            // TOOD: extract boundary.
        }

        /* Try to get boundary if no boundary header
         */
        if (!result.boundary) {
            const boundaryMatch = messageString.match(this.boundaryRegex);

            if (Array.isArray(boundaryMatch) && boundaryMatch.length) {
                fullBoundary = boundaryMatch[0].replace(/[\r\n]+/gi, '');
                const boundary = fullBoundary.replace(/^--/, '');
                result.boundary = boundary;
            }
        }

        /* Parse body
         */
        let start = headerNewlineIndex;

        let contentLength = 0;
        if (result.headers) {
            contentLength = result.headers['Content-Length'] as number;
        }

        let end = (result.headers && result.headers['Content-Length'] && contentLength ? contentLength + start : messageString.length);

        const firstBoundaryIndex = messageString.indexOf(fullBoundary);

        if (firstBoundaryIndex > -1 && result.boundary) {
            start = headerNewlineIndex;
            end = firstBoundaryIndex;
        }

        if (headerNewlineIndex > -1) {
            const body = messageString.slice(start, end);
            result.additional = messageString.slice(end); // Pass back any unparsed data for running thru again
            // console.log("Lengths: total %s -> start %s -> end %s -> final %s", messageString.length, start, end, body.length);

            if (body && body.length) {
                if ((result.headers && result.headers['Content-Type'] === 'application/hap+json') ||
                    (result.headers && result.headers['Content-Type'] === 'application/json')) {
                    // JSON.parse JSON message's
                    try {
                        if (result.headers['Content-Length']) {
                            result.body = body;
                        } else {
                            result.body = body.split('\n')[1];
                        }
                    } catch (err) {

                    }
                } else {
                    result.body = body;
                }
            }
        }

        /* Parse multipart sections
         */
        if (result.boundary) {
            const multipartStart = messageString.indexOf(fullBoundary) + fullBoundary.length;
            const multipartEnd = messageString.lastIndexOf(fullBoundary);
            const multipartBody = messageString.substr(multipartStart, multipartEnd);
            const splitRegex = new RegExp('^' + fullBoundary + '.*[\n\r]?$', 'gm');
            const parts = multipartBody.split(splitRegex);

            result.multipart = parts.filter(this.isTruthy).map((part, index) => {
                // tslint:disable-next-line: no-shadowed-variable
                const result: IHttpResult = {};

                const newlineRegex = /\n\n|\r\n\r\n/gim;
                let newlineIndex = 0;
                let newlineMatch = newlineRegex.exec(part);
                let body = null;

                if (newlineMatch) {
                    newlineIndex = newlineMatch.index;
                    if (newlineMatch.index <= 0) {
                        newlineMatch = newlineRegex.exec(part);
                        if (newlineMatch) {
                            newlineIndex = newlineMatch.index;
                        }
                    }
                }

                const possibleHeadersString = part.substr(0, newlineIndex);

                let startOffset = null;
                let endOffset = null;

                if (newlineIndex > -1) {
                    const headers = this.parseHeaders(possibleHeadersString);
                    if (Object.keys(headers).length > 0) {
                        result.headers = headers;

                        const boundaryIndexes: Array<number> = [];
                        for (let j = 0; j >= 0;) {
                            j = message.indexOf(fullBoundary, j);

                            if (j >= 0) {
                                boundaryIndexes.push(j);
                                j += fullBoundary.length;
                            }
                        }

                        const boundaryNewlineIndexes: Array<number> = [];
                        boundaryIndexes.slice(0, boundaryIndexes.length - 1).forEach(function (m, k) {
                            const partBody = message.slice(boundaryIndexes[k], boundaryIndexes[k + 1]).toString();
                            // tslint:disable-next-line: no-shadowed-variable
                            let headerNewlineIndex = partBody.search(/\n\n|\r\n\r\n/gim) + 2;
                            headerNewlineIndex = boundaryIndexes[k] + headerNewlineIndex;
                            boundaryNewlineIndexes.push(headerNewlineIndex);
                        });

                        startOffset = boundaryNewlineIndexes[index];
                        endOffset = boundaryIndexes[index + 1];
                        body = message.slice(startOffset, endOffset);
                    } else {
                        body = part;
                    }
                } else {
                    body = part;
                }

                result.body = body;
                result.meta.body.byteOffset.start = startOffset;
                result.meta.body.byteOffset.end = endOffset;

                return result;
            });
        }

        return result;

    }

    private trim(str: string): string {
        const firstNonWhitespaceRegex = /[\w-]+/gim;
        const firstNonWhitespaceIndex = str.search(firstNonWhitespaceRegex);
        let tmpStr = "";
        if (firstNonWhitespaceIndex > 0) {
            tmpStr = str.slice(firstNonWhitespaceIndex, str.length);
            str = tmpStr.toString();
        }

        return str;
    }

    private isTruthy = function _isTruthy(value: any): boolean {
        return !!value;
    };

    private isNumeric = function _isNumeric(value: any): boolean {
        if (typeof value === 'number' && !isNaN(value)) {
            return true;
        }

        value = (value || '').toString().trim();

        if (!value) {
            return false;
        }

        return !isNaN(value);
    };

    private isBuffer = (item: any): boolean => {
        return ((this.isNodeBufferSupported() &&
            typeof global === 'object' &&
            global.Buffer.isBuffer(item)) ||
            (item instanceof Object &&
                item._isBuffer));
    };

    private isNodeBufferSupported = (): boolean => {
        return (typeof global === 'object' &&
            typeof global.Buffer === 'function' &&
            typeof global.Buffer.isBuffer === 'function');
    };

    private parseHeaders = (body: any): Dictionary<string | number> => {
        const headers: Dictionary<string | number> = {};

        if (typeof body !== 'string') {
            return headers;
        }

        body.split(/[\r\n]/).forEach((str: string) => {
            const match = str.match(/([\w-]+):\s*(.*)/i);

            if (Array.isArray(match) && match.length === 3) {
                const key = match[1];
                const value = match[2];

                headers[key] = this.isNumeric(value) ? Number(value) : value;
            }
        });

        return headers;
    };
}
